type 'a ab =
	| V
	| N of 'a * 'a ab * 'a ab 

(* nombre de noeuds de a *)
let rec taille (a: 'a ab) : int =
	match a with
	| V -> 0
	| N (x, g, d) -> 1 + taille g + taille d

(* longueur maximale d'un chemin dans a. -1 pour l'arbre vide *)
let rec hauteur (a: 'a ab) : int =
	match a with
	| V -> -1
	| N (x, g, d) -> 1 + max (hauteur g) (hauteur d)

let rec est_feuille (a: 'a ab) = 
	match a with 
	| N(x, V, V) -> true
	| _ -> false

let rec feuilles (a: 'a ab) =
	match a with 
	| V -> 0
	| N(x, V, V) -> 1
	| N(x, g, d) -> feuilles g + feuilles d

let rec etiquette (a: 'a ab) (chemin: bool list) : 'a =
	match (a, chemin) with
	| V, _ -> failwith "chemin invalide" (* pas d'étiquettes dans l'arbre vide *)
	| N(x, g, d), [] -> x 
	| N(x, g, d), true :: q -> etiquette d q
	| N(x, g, d), false :: q -> etiquette g q


let rec etiquette_opt (a: 'a ab) (chemin: bool list) : 'a option =
	match (a, chemin) with
	| V, _ -> None
	| N(x, g, d), [] -> Some x 
	| N(x, g, d), true :: q -> etiquette_opt d q
	| N(x, g, d), false :: q -> etiquette_opt g q

(** PARCOURS **)

let rec print_prefixe (a: int ab) : unit =
	match a with 
	| V -> ()
	| N(x, g, d) -> print_int x; print_prefixe g; print_prefixe d

let rec print_postfixe (a: int ab) : unit =
	match a with 
	| V -> ()
	| N(x, g, d) -> print_postfixe g; print_postfixe d; print_int x

let rec print_infixe (a: int ab) : unit =
	match a with 
	| V -> ()
	| N(x, g, d) -> print_infixe g; print_int x; print_infixe d

let rec taille_add (a: 'a ab) (n: int) : int =
	match a with 
	| V -> n 
	| N(_, g, d) -> 
		let ng = taille_add g (n+1) in 
		taille_add d ng

(* 
Preuve par induction que taille_add a n = taille a + n pour tout (a: 'a ab) et (n: int)

- a = V: taille_add V n = n = taille V + n

- a = N(x, g, d) avec g, d des arbres. 
  On a:
  Notons ng = taille_add g (n+1).
  Alors, par HI, ng = taille g + n + 1
  De plus:
  taille_add a n = taille_add d ng
  par HI, cela vaut
  taille d + ng = taille g + taille g + n + 1
                = taille a + n
D'où le résultat annoncé.
*) 

(* postfixe_concat a l = postfixe a @ l *)
let rec postfixe_concat (a: 'a ab) (l: 'a list) : 'a list =
	match a with 
	| V -> l
	| N(x, g, d) ->
		let ld = postfixe_concat d (x::l) in
		postfixe_concat g ld

let liste_postfixe (a: 'a ab) : 'a list =
	postfixe_concat a []


(** Fonctions pour les listes préfixes et infixes, plus lentes,
	  utilisant la concaténation de liste **)

let rec liste_prefixe (a: 'a ab) : 'a list =
	match a with 
	| V -> []
	| N(x, g, d) -> x :: liste_prefixe g @ liste_prefixe d

(* Renvoie la liste des étiquettes de a dans l'ordre infixe *)
let rec liste_infixe (a: 'a ab) : 'a list =
	match a with 
	| V -> []
	| N(x, g, d) -> liste_infixe g @ x :: liste_infixe d




(*** TESTS ***)
let make_feuille (x: 'a) : 'a ab =
	N(x, V, V)


let tests () =
	let f1 = make_feuille 1 in
	let f2 = make_feuille 2 in
	let f3 = make_feuille 3 in
	let f4 = make_feuille 4 in
	let f5 = make_feuille 5 in
	let f6 = make_feuille 6 in

	let t7 = N(7, f5, f6) in
	let t8 = N(8, f3, f4) in
	let t9 = N(9, t7, t8) in
	let t10 = N(10, f1, f2) in

	let t = N(11, t9, t10) in
	(*
	11-+-9-+-7-+-5
	   |   |   |
	   |   |   +-6
	   |   |
	   |   +-8-+-3
	   |       |
	   |       +-4
	   |
	   +-10-+-1
	        |
	        |
			|
	        +-2
	 *)
	assert (hauteur t = 3);
	assert (hauteur V = -1);
	assert (hauteur t8 = 1);
	assert (hauteur f2 = 0);

	assert (est_feuille f3);
	assert (not (est_feuille t));
	assert (not (est_feuille V));
	assert (not (est_feuille t7));

	assert (feuilles t = 6);
	assert (feuilles f1 = 1);

	assert (etiquette_opt t [false; false] = Some 7);
	assert (etiquette_opt t [false; true; false] = Some 3);
	assert (etiquette_opt t [true; true; false] = None);
	assert (etiquette_opt t [] = Some 11);

	assert (liste_prefixe t = [11; 9; 7; 5; 6; 8; 3; 4; 10; 1; 2]);
	assert (liste_prefixe t8 = [8; 3; 4]);
	assert (liste_prefixe V = []);
	assert (liste_postfixe t = [5; 6; 7; 3; 4; 8; 9; 1; 2; 10; 11]);
	assert (liste_infixe t = [5; 7; 6; 9; 3; 8; 4; 11; 1; 10; 2]);
	
	print_string "Tous les tests ont réussi !\n"