type 'a tree =
	| Node of 'a * 'a tree list (* étiquette, liste des enfants *)


(** Versions plus avancées utilisant des opérations sur les listes
    sans passer par deux fonctions mutuellement récursives **)

(* hauteur a renvoie la profondeur maximale d'une feuille de a,
   -1 pour une liste vide *)
let rec hauteur (Node(x, l): 'a tree) : int =
	1 + List.fold_right max (List.map hauteur l) (-1) 


(* etiquette a c renvoie l'étiquette du noeud de a pointé
   par le chemin c. Échoue avec failwith si c n'est pas un 
   chemin valide dans a *)
let rec etiquette (Node(x, l): 'a tree) (c: int list) : 'a =
	match c with 
	| [] -> x 
	| i :: c' -> 
		let a = List.nth l i in (*List.nth_opt l i renvoie le i-ème élément de l (exception si l est trop courte) *)
		etiquette a c'
		

	(* liste_postfixe a renvoie la liste des étiquettes de a 
   dans l'ordre postfixe *)
let rec liste_prefixe (Node(x, l): 'a tree) : 'a list = 
	x :: List.concat (List.map liste_prefixe l)

(* liste_postfixe_rev a renvoie la liste des étiquettes de a 
   dans l'ordre inverse de l'ordre postfixe *)
let rec liste_postfixe_rev (Node(x, l): 'a tree) : 'a list = 
	x :: List.concat (List.rev (List.map liste_postfixe_rev l))

(* liste_postfixe a renvoie la liste des étiquettes de a 
   dans l'ordre postfixe *)
let liste_postfixe (a: 'a tree) : 'a list = 
	List.rev (liste_postfixe_rev a)


(* Les deux fonctions ci-dessus ne sont pas de complexités linéaires
   car on utilise la concaténation à tout va. Voici une version de
   liste_postfixe qui passe par un accumulateur pour ajouter les
   éléments un par un, sans aucune concaténation *)

(* renvoie liste_postfixe a @ accu *)
let rec postfixe2 (Node(x, l): 'a tree) (accu: 'a list) =
	(* La ligne suivante va partir de x::accu
	   puis y ajouter les éléments du dernier arbre de l
	   puis de l'avant dernier, et ainsi de suite.
	   En effet,
	     List.fold_right postfixe2 [a1; a2; ... an] (x::accu)
	   = postfixe2 a1 (postfixe2 a2 (...  (postfixe2 an (x::accu)) ...)
      = postfixe a1 @ postfixe a2 @ ... @ postfixe an @ x :: accu
      = postfixe a @ accu
	*)
	List.fold_right postfixe2 l (x::accu)



let test () =
	let t = 
	Node(1, [
		Node(2, [
			Node(5, []);
			Node(6, []);
		]);
		Node(3, [
			Node(7, []);
			Node(8, []);
		]);
		Node(4, [
			Node(9, []);
			Node(10, []);
		]);
	])
	in
	assert (liste_prefixe t  = [1; 2; 5; 6; 3; 7; 8; 4; 9; 10]);
	assert (liste_postfixe t = [5; 6; 2; 7; 8; 3; 9; 10; 4; 1]);
	assert (postfixe2 t [] = [5; 6; 2; 7; 8; 3; 9; 10; 4; 1]);

	print_string "Tous les tests ont réussi"