type 'a tree =
	| Node of 'a * 'a tree list (* étiquette, liste des enfants *)


(** Versions avec deux fonctions mutuellement récursives **)

(* hauteur a renvoie la profondeur maximale d'une feuille de a *)
let rec hauteur (Node(x, l): 'a tree) : int =
	1 + hauteur_liste l
(* hauteur_liste l renvoie la hauteur maximale d'un arbre de l,
   -1 si l est vide *)
and hauteur_liste (l: 'a tree list) =
	match l with
	| [] -> -1
	| x :: q -> max (hauteur x) (hauteur_liste q)



(* etiquette a c renvoie l'étiquette du noeud de a pointé
   par le chemin c. Échoue avec failwith si c n'est pas un 
   chemin valide dans a *)
let rec etiquette (Node(x, l): 'a tree) (c: int list) : 'a =
	match c with 
	| [] -> x 
	| i :: c' -> etiquette_liste i l c'
(* etiquette_liste i l c renvoie l'étiquette du noeud pointé 
   par le chemin c dans l'arbre numéro i de l *)
and etiquette_liste (i: int) (l: 'a tree list) (c: int list) : 'a =
	match l, i with
	| [], _ -> failwith "Chemin invalide"
	| a :: q, 0 -> etiquette a c
	| a :: q, _ -> etiquette_liste (i-1) q c 


(* liste_prefixe a renvoie la liste des étiquettes de a 
   dans l'ordre préfixe  *)
let rec liste_prefixe (Node(x, l): 'a tree) : 'a list =
	x :: liste_prefixe_liste l

(* liste_prefixe_liste l renvoie la concaténation des listes
   préfixes de tous les arbres de l *)
and liste_prefixe_liste (l: 'a tree list) : 'a list =
	match l with 
	| [] -> []
	| a :: q -> liste_prefixe a @ liste_prefixe_liste q
	

(** Pour varier un peu: faisons le parcours postfixe avec
    une pile. On a vu que remplaçant la file du parcours
    en largeur par une pile, on retrouve un parcours
    en profondeur. On utilise une fonction
    qui simule un tour de boucle, en dépilant un arbre et empilant
    ses enfants à la place (en faisant attention à l'ordre) **)
let liste_postfixe (a: 'a tree) : 'a list =
	(* boucle pile res renvoie la concaténation des parcours
	   postfixes des arbres de pile, suivi de res *)
	let rec boucle (pile: 'a tree list) (res: 'a list) =
		match pile with 
		| [] -> res
		| Node(x, l) :: q -> boucle (List.rev_append l q) (x :: res)
	in
	boucle [a] []



let tests () =
	let t = 
	Node(1, [
		Node(2, [
			Node(5, []);
			Node(6, []);
		]);
		Node(3, [
			Node(7, []);
			Node(8, []);
		]);
		Node(4, [
			Node(9, []);
			Node(10, []);
		]);
	])
	in
	assert (hauteur t = 2);
	assert (etiquette t [0; 1] = 6);
	assert (etiquette t [] = 1);
	assert (liste_prefixe t  = [1; 2; 5; 6; 3; 7; 8; 4; 9; 10]);
	assert (liste_postfixe t = [5; 6; 2; 7; 8; 3; 9; 10; 4; 1]);
	print_string "Tous les tests ont réussi\n"